import { Map, OrderedMap, OrderedSet, Set, Collection } from "immutable";
import { PrimitiveTypeKind, Type, PrimitiveStringTypeKind, ClassProperty, TypeKind } from "./Type";
import { TypeGraph } from "./TypeGraph";
import { TypeAttributes, TypeAttributeKind } from "./TypeAttributes";
export declare type TypeRefCallback = (index: number) => void;
export declare class TypeRef {
    readonly graph: TypeGraph;
    readonly maybeIndex: number;
    constructor(graph: TypeGraph, maybeIndex: number);
    getIndex(): number;
    callWhenResolved(callback: TypeRefCallback): void;
    deref(): [Type, TypeAttributes];
    equals(other: any): boolean;
    hashCode(): number;
}
export declare const provenanceTypeAttributeKind: TypeAttributeKind<Set<TypeRef>>;
export declare type StringTypeMapping = {
    date: PrimitiveStringTypeKind;
    time: PrimitiveStringTypeKind;
    dateTime: PrimitiveStringTypeKind;
};
export declare const NoStringTypeMapping: StringTypeMapping;
export declare class TypeBuilder {
    private readonly _stringTypeMapping;
    readonly alphabetizeProperties: boolean;
    private readonly _allPropertiesOptional;
    private readonly _addProvenanceAttributes;
    readonly typeGraph: TypeGraph;
    protected topLevels: Map<string, TypeRef>;
    protected readonly types: (Type | undefined)[];
    private readonly typeAttributes;
    private _addedForwardingIntersection;
    constructor(_stringTypeMapping: StringTypeMapping, alphabetizeProperties: boolean, _allPropertiesOptional: boolean, _addProvenanceAttributes: boolean, inheritsProvenanceAttributes: boolean);
    addTopLevel(name: string, tref: TypeRef): void;
    reserveTypeRef(): TypeRef;
    private commitType;
    protected addType<T extends Type>(forwardingRef: TypeRef | undefined, creator: (tref: TypeRef) => T, attributes: TypeAttributes | undefined): TypeRef;
    atIndex(index: number): [Type, TypeAttributes];
    addAttributes(tref: TypeRef, attributes: TypeAttributes | undefined): void;
    makeNullable(tref: TypeRef, attributes: TypeAttributes): TypeRef;
    finish(): TypeGraph;
    protected addForwardingIntersection(forwardingRef: TypeRef, tref: TypeRef): TypeRef;
    protected forwardIfNecessary(forwardingRef: TypeRef | undefined, tref: undefined): undefined;
    protected forwardIfNecessary(forwardingRef: TypeRef | undefined, tref: TypeRef): TypeRef;
    protected forwardIfNecessary(forwardingRef: TypeRef | undefined, tref: TypeRef | undefined): TypeRef | undefined;
    readonly didAddForwardingIntersection: boolean;
    private _primitiveTypes;
    private _noEnumStringType;
    private _mapTypes;
    private _arrayTypes;
    private _enumTypes;
    private _classTypes;
    private _unionTypes;
    private _intersectionTypes;
    getPrimitiveType(kind: PrimitiveTypeKind, forwardingRef?: TypeRef): TypeRef;
    getStringType(attributes: TypeAttributes | undefined, cases: OrderedMap<string, number> | undefined, forwardingRef?: TypeRef): TypeRef;
    getEnumType(attributes: TypeAttributes, cases: OrderedSet<string>, forwardingRef?: TypeRef): TypeRef;
    getUniqueObjectType(attributes: TypeAttributes, properties: OrderedMap<string, ClassProperty> | undefined, additionalProperties: TypeRef | undefined, forwardingRef?: TypeRef): TypeRef;
    getUniqueMapType(forwardingRef?: TypeRef): TypeRef;
    private registerMapType(values, tref);
    getMapType(values: TypeRef, forwardingRef?: TypeRef): TypeRef;
    private registerClassType(properties, tref);
    setObjectProperties(ref: TypeRef, properties: OrderedMap<string, ClassProperty>, additionalProperties: TypeRef | undefined): void;
    getUniqueArrayType(forwardingRef?: TypeRef): TypeRef;
    private registerArrayType(items, tref);
    getArrayType(items: TypeRef, forwardingRef?: TypeRef): TypeRef;
    setArrayItems(ref: TypeRef, items: TypeRef): void;
    modifyPropertiesIfNecessary(properties: OrderedMap<string, ClassProperty>): OrderedMap<string, ClassProperty>;
    getClassType(attributes: TypeAttributes, properties: OrderedMap<string, ClassProperty>, forwardingRef?: TypeRef): TypeRef;
    private registerUnionType(members, tref);
    getUniqueClassType(attributes: TypeAttributes, isFixed: boolean, properties: OrderedMap<string, ClassProperty> | undefined, forwardingRef?: TypeRef): TypeRef;
    getUnionType(attributes: TypeAttributes, members: OrderedSet<TypeRef>, forwardingRef?: TypeRef): TypeRef;
    getUniqueUnionType(attributes: TypeAttributes, members: OrderedSet<TypeRef> | undefined, forwardingRef?: TypeRef): TypeRef;
    private registerIntersectionType(members, tref);
    getIntersectionType(attributes: TypeAttributes, members: OrderedSet<TypeRef>, forwardingRef?: TypeRef): TypeRef;
    getUniqueIntersectionType(attributes: TypeAttributes, members: OrderedSet<TypeRef> | undefined, forwardingRef?: TypeRef): TypeRef;
    setSetOperationMembers(ref: TypeRef, members: OrderedSet<TypeRef>): void;
    setLostTypeAttributes(): void;
}
export interface TypeLookerUp {
    lookupTypeRefs(typeRefs: TypeRef[], forwardingRef?: TypeRef): TypeRef | undefined;
    reconstituteTypeRef(typeRef: TypeRef, forwardingRef?: TypeRef): TypeRef;
}
export declare class TypeReconstituter<TBuilder extends BaseGraphRewriteBuilder> {
    private readonly _typeBuilder;
    private readonly _makeClassUnique;
    private readonly _typeAttributes;
    private readonly _forwardingRef;
    private readonly _register;
    private _wasUsed;
    private _typeRef;
    constructor(_typeBuilder: TBuilder, _makeClassUnique: boolean, _typeAttributes: TypeAttributes, _forwardingRef: TypeRef | undefined, _register: (tref: TypeRef) => void);
    private builderForNewType();
    private builderForSetting();
    getResult(): TypeRef;
    private register(tref);
    private registerAndAddAttributes(tref);
    lookup(tref: TypeRef): TypeRef | undefined;
    lookup<C extends Collection<any, TypeRef>>(trefs: C): C | undefined;
    reconstitute(tref: TypeRef): TypeRef;
    reconstitute<C extends Collection<any, TypeRef>>(trefs: C): C;
    getPrimitiveType(kind: PrimitiveTypeKind): void;
    getStringType(enumCases: OrderedMap<string, number> | undefined): void;
    getEnumType(cases: OrderedSet<string>): void;
    getUniqueMapType(): void;
    getMapType(values: TypeRef): void;
    getUniqueArrayType(): void;
    getArrayType(items: TypeRef): void;
    setArrayItems(items: TypeRef): void;
    getObjectType(properties: OrderedMap<string, ClassProperty>, additionalProperties: TypeRef | undefined): void;
    getUniqueObjectType(properties: OrderedMap<string, ClassProperty> | undefined, additionalProperties: TypeRef | undefined): void;
    getClassType(properties: OrderedMap<string, ClassProperty>): void;
    getUniqueClassType(isFixed: boolean, properties: OrderedMap<string, ClassProperty> | undefined): void;
    setObjectProperties(properties: OrderedMap<string, ClassProperty>, additionalProperties: TypeRef | undefined): void;
    getUnionType(members: OrderedSet<TypeRef>): void;
    getUniqueUnionType(): void;
    getIntersectionType(members: OrderedSet<TypeRef>): void;
    getUniqueIntersectionType(members?: OrderedSet<TypeRef>): void;
    setSetOperationMembers(members: OrderedSet<TypeRef>): void;
}
export declare abstract class BaseGraphRewriteBuilder extends TypeBuilder implements TypeLookerUp {
    protected readonly originalGraph: TypeGraph;
    protected readonly debugPrint: boolean;
    protected reconstitutedTypes: Map<number, TypeRef>;
    private _lostTypeAttributes;
    private _printIndent;
    constructor(originalGraph: TypeGraph, stringTypeMapping: StringTypeMapping, alphabetizeProperties: boolean, graphHasProvenanceAttributes: boolean, debugPrint: boolean);
    reconstituteType(t: Type, forwardingRef?: TypeRef): TypeRef;
    abstract lookupTypeRefs(typeRefs: TypeRef[], forwardingRef?: TypeRef, replaceSet?: boolean): TypeRef | undefined;
    protected abstract forceReconstituteTypeRef(originalRef: TypeRef, maybeForwardingRef?: TypeRef): TypeRef;
    reconstituteTypeRef(originalRef: TypeRef, maybeForwardingRef?: TypeRef): TypeRef;
    protected assertTypeRefsToReconstitute(typeRefs: TypeRef[], forwardingRef?: TypeRef): void;
    protected changeDebugPrintIndent(delta: number): void;
    protected readonly debugPrintIndentation: string;
    finish(): TypeGraph;
    setLostTypeAttributes(): void;
    readonly lostTypeAttributes: boolean;
}
export declare class GraphRemapBuilder extends BaseGraphRewriteBuilder {
    private readonly _map;
    private _attributeSources;
    constructor(originalGraph: TypeGraph, stringTypeMapping: StringTypeMapping, alphabetizeProperties: boolean, graphHasProvenanceAttributes: boolean, _map: Map<Type, Type>, debugPrintRemapping: boolean);
    private getMapTarget(tref);
    protected addForwardingIntersection(_forwardingRef: TypeRef, _tref: TypeRef): TypeRef;
    lookupTypeRefs(typeRefs: TypeRef[], forwardingRef?: TypeRef): TypeRef | undefined;
    protected forceReconstituteTypeRef(originalRef: TypeRef, maybeForwardingRef?: TypeRef): TypeRef;
}
export declare class GraphRewriteBuilder<T extends Type> extends BaseGraphRewriteBuilder {
    private readonly _replacer;
    private _setsToReplaceByMember;
    private _reconstitutedUnions;
    constructor(originalGraph: TypeGraph, stringTypeMapping: StringTypeMapping, alphabetizeProperties: boolean, graphHasProvenanceAttributes: boolean, setsToReplace: T[][], debugPrintReconstitution: boolean, _replacer: (typesToReplace: Set<T>, builder: GraphRewriteBuilder<T>, forwardingRef: TypeRef) => TypeRef);
    registerUnion(typeRefs: TypeRef[], reconstituted: TypeRef): void;
    withForwardingRef(maybeForwardingRef: TypeRef | undefined, typeCreator: (forwardingRef: TypeRef) => TypeRef): TypeRef;
    private replaceSet(typesToReplace, maybeForwardingRef);
    protected forceReconstituteTypeRef(originalRef: TypeRef, maybeForwardingRef?: TypeRef): TypeRef;
    lookupTypeRefs(typeRefs: TypeRef[], forwardingRef?: TypeRef, replaceSet?: boolean): TypeRef | undefined;
}
export interface UnionTypeProvider<TArrayData, TObjectData> {
    readonly arrayData: TArrayData;
    readonly objectData: TObjectData;
    enumCaseMap: {
        [name: string]: number;
    };
    enumCases: string[];
    getMemberKinds(): TypeAttributeMap<TypeKind>;
    readonly lostTypeAttributes: boolean;
}
export declare type TypeAttributeMap<T extends TypeKind> = OrderedMap<T, TypeAttributes>;
export declare class UnionAccumulator<TArray, TObject> implements UnionTypeProvider<TArray[], TObject[]> {
    private readonly _conflateNumbers;
    private _nonStringTypeAttributes;
    private _stringTypeAttributes;
    readonly arrayData: TArray[];
    readonly objectData: TObject[];
    private _lostTypeAttributes;
    enumCaseMap: {
        [name: string]: number;
    };
    enumCases: string[];
    constructor(_conflateNumbers: boolean);
    private have(kind);
    readonly haveString: boolean;
    addAny(attributes: TypeAttributes): void;
    addNull(attributes: TypeAttributes): void;
    addBool(attributes: TypeAttributes): void;
    addInteger(attributes: TypeAttributes): void;
    addDouble(attributes: TypeAttributes): void;
    addStringType(kind: PrimitiveStringTypeKind, attributes: TypeAttributes): void;
    addArray(t: TArray, attributes: TypeAttributes): void;
    addObject(t: TObject, attributes: TypeAttributes): void;
    addEnumCases(cases: OrderedMap<string, number>, attributes: TypeAttributes): void;
    addEnumCase(s: string, count: number, attributes: TypeAttributes): void;
    getMemberKinds(): TypeAttributeMap<TypeKind>;
    readonly lostTypeAttributes: boolean;
}
export declare class TypeRefUnionAccumulator extends UnionAccumulator<TypeRef, TypeRef> {
    private addType(t, attributes);
    addTypes(types: Set<Type>): TypeAttributes;
}
export declare abstract class UnionBuilder<TBuilder extends TypeBuilder, TArrayData, TObjectData> {
    protected readonly typeBuilder: TBuilder;
    constructor(typeBuilder: TBuilder);
    protected abstract makeEnum(cases: string[], counts: {
        [name: string]: number;
    }, typeAttributes: TypeAttributes, forwardingRef: TypeRef | undefined): TypeRef;
    protected abstract makeObject(objects: TObjectData, typeAttributes: TypeAttributes, forwardingRef: TypeRef | undefined): TypeRef;
    protected abstract makeArray(arrays: TArrayData, typeAttributes: TypeAttributes, forwardingRef: TypeRef | undefined): TypeRef;
    private makeTypeOfKind(typeProvider, kind, typeAttributes, forwardingRef);
    buildUnion(typeProvider: UnionTypeProvider<TArrayData, TObjectData>, unique: boolean, typeAttributes: TypeAttributes, forwardingRef?: TypeRef): TypeRef;
}
