"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Type_1 = require("./Type");
var TypeGraph_1 = require("./TypeGraph");
var TypeAttributes_1 = require("./TypeAttributes");
var Support_1 = require("./Support");
// FIXME: Get rid of this eventually
var TypeRef = /** @class */ (function () {
    function TypeRef(graph, maybeIndex) {
        this.graph = graph;
        this.maybeIndex = maybeIndex;
    }
    TypeRef.prototype.getIndex = function () {
        return this.maybeIndex;
    };
    TypeRef.prototype.callWhenResolved = function (callback) {
        callback(this.getIndex());
    };
    TypeRef.prototype.deref = function () {
        return this.graph.atIndex(this.getIndex());
    };
    TypeRef.prototype.equals = function (other) {
        if (!(other instanceof TypeRef)) {
            return false;
        }
        Support_1.assert(this.graph === other.graph, "Comparing type refs of different graphs");
        return this.maybeIndex === other.maybeIndex;
    };
    TypeRef.prototype.hashCode = function () {
        return this.maybeIndex | 0;
    };
    return TypeRef;
}());
exports.TypeRef = TypeRef;
function provenanceToString(p) {
    return p
        .map(function (r) { return r.getIndex(); })
        .toList()
        .sort()
        .map(function (i) { return i.toString(); })
        .join(",");
}
// FIXME: Don't infer provenance.  All original types should be present in
// non-inferred form in the final graph.
exports.provenanceTypeAttributeKind = new TypeAttributes_1.TypeAttributeKind("provenance", Support_1.setUnion, function (a) { return a; }, provenanceToString);
exports.NoStringTypeMapping = {
    date: "date",
    time: "time",
    dateTime: "date-time"
};
var TypeBuilder = /** @class */ (function () {
    function TypeBuilder(_stringTypeMapping, alphabetizeProperties, _allPropertiesOptional, _addProvenanceAttributes, inheritsProvenanceAttributes) {
        var _this = this;
        this._stringTypeMapping = _stringTypeMapping;
        this.alphabetizeProperties = alphabetizeProperties;
        this._allPropertiesOptional = _allPropertiesOptional;
        this._addProvenanceAttributes = _addProvenanceAttributes;
        this.topLevels = immutable_1.Map();
        this.types = [];
        this.typeAttributes = [];
        this._addedForwardingIntersection = false;
        this.commitType = function (tref, t) {
            var index = tref.getIndex();
            // const name = names !== undefined ? ` ${names.combinedName}` : "";
            // console.log(`committing ${t.kind}${name} to ${index}`);
            Support_1.assert(_this.types[index] === undefined, "A type index was committed twice");
            _this.types[index] = t;
        };
        // FIXME: make mutable?
        this._primitiveTypes = immutable_1.Map();
        this._noEnumStringType = undefined;
        // FIXME: Combine map and object type maps
        this._mapTypes = immutable_1.Map();
        this._arrayTypes = immutable_1.Map();
        this._enumTypes = immutable_1.Map();
        this._classTypes = immutable_1.Map();
        this._unionTypes = immutable_1.Map();
        this._intersectionTypes = immutable_1.Map();
        Support_1.assert(!_addProvenanceAttributes || !inheritsProvenanceAttributes, "We can't both inherit as well as add provenance");
        this.typeGraph = new TypeGraph_1.TypeGraph(this, _addProvenanceAttributes || inheritsProvenanceAttributes);
    }
    TypeBuilder.prototype.addTopLevel = function (name, tref) {
        // assert(t.typeGraph === this.typeGraph, "Adding top-level to wrong type graph");
        Support_1.assert(!this.topLevels.has(name), "Trying to add top-level with existing name");
        Support_1.assert(this.types[tref.getIndex()] !== undefined, "Trying to add a top-level type that doesn't exist (yet?)");
        this.topLevels = this.topLevels.set(name, tref);
    };
    TypeBuilder.prototype.reserveTypeRef = function () {
        var index = this.types.length;
        // console.log(`reserving ${index}`);
        this.types.push(undefined);
        var tref = new TypeRef(this.typeGraph, index);
        var attributes = this._addProvenanceAttributes
            ? exports.provenanceTypeAttributeKind.makeAttributes(immutable_1.Set([tref]))
            : immutable_1.Map();
        this.typeAttributes.push(attributes);
        return tref;
    };
    TypeBuilder.prototype.addType = function (forwardingRef, creator, attributes) {
        if (forwardingRef !== undefined && forwardingRef.maybeIndex !== undefined) {
            Support_1.assert(this.types[forwardingRef.maybeIndex] === undefined);
        }
        var tref = forwardingRef !== undefined ? forwardingRef : this.reserveTypeRef();
        if (attributes !== undefined) {
            this.addAttributes(tref, attributes);
        }
        var t = creator(tref);
        this.commitType(tref, t);
        return tref;
    };
    TypeBuilder.prototype.atIndex = function (index) {
        var maybeType = this.types[index];
        if (maybeType === undefined) {
            return Support_1.panic("Trying to deref an undefined type in a type builder");
        }
        var maybeNames = this.typeAttributes[index];
        return [maybeType, maybeNames];
    };
    TypeBuilder.prototype.addAttributes = function (tref, attributes) {
        var _this = this;
        tref.callWhenResolved(function (index) {
            if (attributes === undefined) {
                attributes = immutable_1.Map();
            }
            _this.typeAttributes[index] = TypeAttributes_1.combineTypeAttributes(_this.typeAttributes[index], attributes);
        });
    };
    TypeBuilder.prototype.makeNullable = function (tref, attributes) {
        var t = Support_1.defined(this.types[tref.getIndex()]);
        if (t.kind === "null" || t.kind === "any") {
            return tref;
        }
        var nullType = this.getPrimitiveType("null");
        if (!(t instanceof Type_1.UnionType)) {
            return this.getUnionType(attributes, immutable_1.OrderedSet([tref, nullType]));
        }
        var _a = Type_1.removeNullFromUnion(t), maybeNull = _a[0], nonNulls = _a[1];
        if (maybeNull !== null)
            return tref;
        return this.getUnionType(attributes, nonNulls.map(function (nn) { return nn.typeRef; }).add(nullType));
    };
    TypeBuilder.prototype.finish = function () {
        this.typeGraph.freeze(this.topLevels, this.types.map(Support_1.defined), this.typeAttributes);
        return this.typeGraph;
    };
    TypeBuilder.prototype.addForwardingIntersection = function (forwardingRef, tref) {
        this._addedForwardingIntersection = true;
        return this.addType(forwardingRef, function (tr) { return new Type_1.IntersectionType(tr, immutable_1.OrderedSet([tref])); }, undefined);
    };
    TypeBuilder.prototype.forwardIfNecessary = function (forwardingRef, tref) {
        if (tref === undefined)
            return undefined;
        if (forwardingRef === undefined)
            return tref;
        return this.addForwardingIntersection(forwardingRef, tref);
    };
    Object.defineProperty(TypeBuilder.prototype, "didAddForwardingIntersection", {
        get: function () {
            return this._addedForwardingIntersection;
        },
        enumerable: true,
        configurable: true
    });
    TypeBuilder.prototype.getPrimitiveType = function (kind, forwardingRef) {
        Support_1.assert(kind !== "string", "Use getStringType to create strings");
        if (kind === "date")
            kind = this._stringTypeMapping.date;
        if (kind === "time")
            kind = this._stringTypeMapping.time;
        if (kind === "date-time")
            kind = this._stringTypeMapping.dateTime;
        if (kind === "string") {
            return this.getStringType(undefined, undefined, forwardingRef);
        }
        var tref = this.forwardIfNecessary(forwardingRef, this._primitiveTypes.get(kind));
        if (tref === undefined) {
            tref = this.addType(forwardingRef, function (tr) { return new Type_1.PrimitiveType(tr, kind); }, undefined);
            this._primitiveTypes = this._primitiveTypes.set(kind, tref);
        }
        return tref;
    };
    TypeBuilder.prototype.getStringType = function (attributes, cases, forwardingRef) {
        if (cases === undefined) {
            // FIXME: Right now we completely ignore names for strings
            // without enum cases.  That's the correct behavior at the time,
            // because string types never are assigned names, but we might
            // do that at some point, but in that case we'll want a different
            // type for each occurrence, not the same single string type with
            // all the names.
            //
            // The proper solution at that point might be to just figure
            // out whether we do want string types to have names (we most
            // likely don't), and if not, still don't keep track of them.
            var result = void 0;
            if (this._noEnumStringType === undefined) {
                result = this._noEnumStringType = this.addType(forwardingRef, function (tr) { return new Type_1.StringType(tr, undefined); }, undefined);
            }
            else {
                result = this.forwardIfNecessary(forwardingRef, this._noEnumStringType);
            }
            this.addAttributes(this._noEnumStringType, attributes);
            return result;
        }
        return this.addType(forwardingRef, function (tr) { return new Type_1.StringType(tr, cases); }, attributes);
    };
    TypeBuilder.prototype.getEnumType = function (attributes, cases, forwardingRef) {
        var unorderedCases = cases.toSet();
        var tref = this.forwardIfNecessary(forwardingRef, this._enumTypes.get(unorderedCases));
        if (tref === undefined) {
            tref = this.addType(forwardingRef, function (tr) { return new Type_1.EnumType(tr, cases); }, attributes);
            this._enumTypes = this._enumTypes.set(unorderedCases, tref);
        }
        else {
            this.addAttributes(tref, attributes);
        }
        return tref;
    };
    TypeBuilder.prototype.getUniqueObjectType = function (attributes, properties, additionalProperties, forwardingRef) {
        var _this = this;
        properties = Support_1.mapOptional(function (p) { return _this.modifyPropertiesIfNecessary(p); }, properties);
        return this.addType(forwardingRef, function (tref) { return new Type_1.ObjectType(tref, "object", true, properties, additionalProperties); }, attributes);
    };
    TypeBuilder.prototype.getUniqueMapType = function (forwardingRef) {
        return this.addType(forwardingRef, function (tr) { return new Type_1.MapType(tr, undefined); }, undefined);
    };
    TypeBuilder.prototype.registerMapType = function (values, tref) {
        if (this._mapTypes.has(values))
            return;
        this._mapTypes = this._mapTypes.set(values, tref);
    };
    TypeBuilder.prototype.getMapType = function (values, forwardingRef) {
        var tref = this.forwardIfNecessary(forwardingRef, this._mapTypes.get(values));
        if (tref === undefined) {
            tref = this.addType(forwardingRef, function (tr) { return new Type_1.MapType(tr, values); }, undefined);
            this.registerMapType(values, tref);
        }
        return tref;
    };
    TypeBuilder.prototype.registerClassType = function (properties, tref) {
        var map = properties.toMap();
        this._classTypes = this._classTypes.set(map, tref);
    };
    TypeBuilder.prototype.setObjectProperties = function (ref, properties, additionalProperties) {
        var type = ref.deref()[0];
        if (!(type instanceof Type_1.ObjectType)) {
            return Support_1.panic("Tried to set properties of non-object type");
        }
        type.setProperties(this.modifyPropertiesIfNecessary(properties), additionalProperties);
        if (type instanceof Type_1.MapType) {
            this.registerMapType(Support_1.defined(additionalProperties), ref);
        }
        else if (type instanceof Type_1.ClassType) {
            if (!type.isFixed) {
                this.registerClassType(properties, ref);
            }
        }
    };
    TypeBuilder.prototype.getUniqueArrayType = function (forwardingRef) {
        return this.addType(forwardingRef, function (tr) { return new Type_1.ArrayType(tr, undefined); }, undefined);
    };
    TypeBuilder.prototype.registerArrayType = function (items, tref) {
        if (this._arrayTypes.has(items))
            return;
        this._arrayTypes = this._arrayTypes.set(items, tref);
    };
    TypeBuilder.prototype.getArrayType = function (items, forwardingRef) {
        var tref = this.forwardIfNecessary(forwardingRef, this._arrayTypes.get(items));
        if (tref === undefined) {
            tref = this.addType(forwardingRef, function (tr) { return new Type_1.ArrayType(tr, items); }, undefined);
            this.registerArrayType(items, tref);
        }
        return tref;
    };
    TypeBuilder.prototype.setArrayItems = function (ref, items) {
        var type = ref.deref()[0];
        if (!(type instanceof Type_1.ArrayType)) {
            return Support_1.panic("Tried to set items of non-array type");
        }
        type.setItems(items);
        this.registerArrayType(items, ref);
    };
    TypeBuilder.prototype.modifyPropertiesIfNecessary = function (properties) {
        if (this.alphabetizeProperties) {
            properties = properties.sortBy(function (_, n) { return n; });
        }
        if (this._allPropertiesOptional) {
            properties = properties.map(function (cp) { return new Type_1.ClassProperty(cp.typeRef, true); });
        }
        return properties;
    };
    TypeBuilder.prototype.getClassType = function (attributes, properties, forwardingRef) {
        properties = this.modifyPropertiesIfNecessary(properties);
        var tref = this.forwardIfNecessary(forwardingRef, this._classTypes.get(properties.toMap()));
        if (tref === undefined) {
            tref = this.addType(forwardingRef, function (tr) { return new Type_1.ClassType(tr, false, properties); }, attributes);
            this.registerClassType(properties, tref);
        }
        else {
            this.addAttributes(tref, attributes);
        }
        return tref;
    };
    TypeBuilder.prototype.registerUnionType = function (members, tref) {
        if (this._unionTypes.has(members))
            return;
        this._unionTypes = this._unionTypes.set(members, tref);
    };
    // FIXME: Maybe just distinguish between this and `getClassType`
    // via a flag?  That would make `ClassType.map` simpler.
    TypeBuilder.prototype.getUniqueClassType = function (attributes, isFixed, properties, forwardingRef) {
        var _this = this;
        properties = Support_1.mapOptional(function (p) { return _this.modifyPropertiesIfNecessary(p); }, properties);
        return this.addType(forwardingRef, function (tref) { return new Type_1.ClassType(tref, isFixed, properties); }, attributes);
    };
    TypeBuilder.prototype.getUnionType = function (attributes, members, forwardingRef) {
        var unorderedMembers = members.toSet();
        var tref = this.forwardIfNecessary(forwardingRef, this._unionTypes.get(unorderedMembers));
        if (tref === undefined) {
            tref = this.addType(forwardingRef, function (tr) { return new Type_1.UnionType(tr, members); }, attributes);
            this.registerUnionType(unorderedMembers, tref);
        }
        else {
            this.addAttributes(tref, attributes);
        }
        return tref;
    };
    TypeBuilder.prototype.getUniqueUnionType = function (attributes, members, forwardingRef) {
        return this.addType(forwardingRef, function (tref) { return new Type_1.UnionType(tref, members); }, attributes);
    };
    TypeBuilder.prototype.registerIntersectionType = function (members, tref) {
        if (this._intersectionTypes.has(members))
            return;
        this._intersectionTypes = this._intersectionTypes.set(members, tref);
    };
    TypeBuilder.prototype.getIntersectionType = function (attributes, members, forwardingRef) {
        var unorderedMembers = members.toSet();
        var tref = this.forwardIfNecessary(forwardingRef, this._intersectionTypes.get(unorderedMembers));
        if (tref === undefined) {
            tref = this.addType(forwardingRef, function (tr) { return new Type_1.IntersectionType(tr, members); }, attributes);
            this.registerIntersectionType(unorderedMembers, tref);
        }
        else {
            this.addAttributes(tref, attributes);
        }
        return tref;
    };
    TypeBuilder.prototype.getUniqueIntersectionType = function (attributes, members, forwardingRef) {
        return this.addType(forwardingRef, function (tref) { return new Type_1.IntersectionType(tref, members); }, attributes);
    };
    TypeBuilder.prototype.setSetOperationMembers = function (ref, members) {
        var type = ref.deref()[0];
        if (!(type instanceof Type_1.UnionType || type instanceof Type_1.IntersectionType)) {
            return Support_1.panic("Tried to set members of non-set-operation type");
        }
        type.setMembers(members);
        var unorderedMembers = members.toSet();
        if (type instanceof Type_1.UnionType) {
            this.registerUnionType(unorderedMembers, ref);
        }
        else if (type instanceof Type_1.IntersectionType) {
            this.registerIntersectionType(unorderedMembers, ref);
        }
        else {
            return Support_1.panic("Unknown set operation type ${type.kind}");
        }
    };
    TypeBuilder.prototype.setLostTypeAttributes = function () {
        return;
    };
    return TypeBuilder;
}());
exports.TypeBuilder = TypeBuilder;
var TypeReconstituter = /** @class */ (function () {
    function TypeReconstituter(_typeBuilder, _makeClassUnique, _typeAttributes, _forwardingRef, _register) {
        this._typeBuilder = _typeBuilder;
        this._makeClassUnique = _makeClassUnique;
        this._typeAttributes = _typeAttributes;
        this._forwardingRef = _forwardingRef;
        this._register = _register;
        this._wasUsed = false;
        this._typeRef = undefined;
    }
    TypeReconstituter.prototype.builderForNewType = function () {
        Support_1.assert(!this._wasUsed, "TypeReconstituter used more than once");
        this._wasUsed = true;
        return this._typeBuilder;
    };
    TypeReconstituter.prototype.builderForSetting = function () {
        Support_1.assert(this._wasUsed && this._typeRef !== undefined, "Can't set type members before constructing a type");
        return this._typeBuilder;
    };
    TypeReconstituter.prototype.getResult = function () {
        if (this._typeRef === undefined) {
            return Support_1.panic("Type was not reconstituted");
        }
        return this._typeRef;
    };
    // FIXME: Do registration automatically.
    TypeReconstituter.prototype.register = function (tref) {
        Support_1.assert(this._typeRef === undefined, "Cannot register a type twice");
        this._typeRef = tref;
        this._register(tref);
    };
    TypeReconstituter.prototype.registerAndAddAttributes = function (tref) {
        this._typeBuilder.addAttributes(tref, this._typeAttributes);
        this.register(tref);
    };
    TypeReconstituter.prototype.lookup = function (trefs) {
        var _this = this;
        Support_1.assert(!this._wasUsed, "Cannot lookup constituents after building type");
        if (immutable_1.isCollection(trefs)) {
            var maybeRefs = trefs.map(function (tref) { return _this._typeBuilder.lookupTypeRefs([tref], undefined, false); });
            if (maybeRefs.some(function (tref) { return tref === undefined; }))
                return undefined;
            return maybeRefs;
        }
        return this._typeBuilder.lookupTypeRefs([trefs], undefined, false);
    };
    TypeReconstituter.prototype.reconstitute = function (trefs) {
        var _this = this;
        Support_1.assert(this._wasUsed, "Cannot reconstitute constituents before building type");
        if (immutable_1.isCollection(trefs)) {
            return trefs.map(function (tref) { return _this._typeBuilder.reconstituteTypeRef(tref); });
        }
        return this._typeBuilder.reconstituteTypeRef(trefs);
    };
    TypeReconstituter.prototype.getPrimitiveType = function (kind) {
        this.registerAndAddAttributes(this.builderForNewType().getPrimitiveType(kind, this._forwardingRef));
    };
    TypeReconstituter.prototype.getStringType = function (enumCases) {
        this.register(this.builderForNewType().getStringType(this._typeAttributes, enumCases, this._forwardingRef));
    };
    TypeReconstituter.prototype.getEnumType = function (cases) {
        this.register(this.builderForNewType().getEnumType(this._typeAttributes, cases, this._forwardingRef));
    };
    TypeReconstituter.prototype.getUniqueMapType = function () {
        this.registerAndAddAttributes(this.builderForNewType().getUniqueMapType(this._forwardingRef));
    };
    TypeReconstituter.prototype.getMapType = function (values) {
        this.registerAndAddAttributes(this.builderForNewType().getMapType(values, this._forwardingRef));
    };
    TypeReconstituter.prototype.getUniqueArrayType = function () {
        this.registerAndAddAttributes(this.builderForNewType().getUniqueArrayType(this._forwardingRef));
    };
    TypeReconstituter.prototype.getArrayType = function (items) {
        this.registerAndAddAttributes(this.builderForNewType().getArrayType(items, this._forwardingRef));
    };
    TypeReconstituter.prototype.setArrayItems = function (items) {
        this.builderForSetting().setArrayItems(this.getResult(), items);
    };
    TypeReconstituter.prototype.getObjectType = function (properties, additionalProperties) {
        this.register(this.builderForNewType().getUniqueObjectType(this._typeAttributes, properties, additionalProperties, this._forwardingRef));
    };
    TypeReconstituter.prototype.getUniqueObjectType = function (properties, additionalProperties) {
        this.register(this.builderForNewType().getUniqueObjectType(this._typeAttributes, properties, additionalProperties, this._forwardingRef));
    };
    TypeReconstituter.prototype.getClassType = function (properties) {
        if (this._makeClassUnique) {
            this.getUniqueClassType(false, properties);
            return;
        }
        this.register(this.builderForNewType().getClassType(this._typeAttributes, properties, this._forwardingRef));
    };
    TypeReconstituter.prototype.getUniqueClassType = function (isFixed, properties) {
        this.register(this.builderForNewType().getUniqueClassType(this._typeAttributes, isFixed, properties, this._forwardingRef));
    };
    TypeReconstituter.prototype.setObjectProperties = function (properties, additionalProperties) {
        this.builderForSetting().setObjectProperties(this.getResult(), properties, additionalProperties);
    };
    TypeReconstituter.prototype.getUnionType = function (members) {
        this.register(this.builderForNewType().getUnionType(this._typeAttributes, members, this._forwardingRef));
    };
    TypeReconstituter.prototype.getUniqueUnionType = function () {
        this.register(this.builderForNewType().getUniqueUnionType(this._typeAttributes, undefined, this._forwardingRef));
    };
    TypeReconstituter.prototype.getIntersectionType = function (members) {
        this.register(this.builderForNewType().getIntersectionType(this._typeAttributes, members, this._forwardingRef));
    };
    TypeReconstituter.prototype.getUniqueIntersectionType = function (members) {
        this.register(this.builderForNewType().getUniqueIntersectionType(this._typeAttributes, members, this._forwardingRef));
    };
    TypeReconstituter.prototype.setSetOperationMembers = function (members) {
        this.builderForSetting().setSetOperationMembers(this.getResult(), members);
    };
    return TypeReconstituter;
}());
exports.TypeReconstituter = TypeReconstituter;
var BaseGraphRewriteBuilder = /** @class */ (function (_super) {
    __extends(BaseGraphRewriteBuilder, _super);
    function BaseGraphRewriteBuilder(originalGraph, stringTypeMapping, alphabetizeProperties, graphHasProvenanceAttributes, debugPrint) {
        var _this = _super.call(this, stringTypeMapping, alphabetizeProperties, false, false, graphHasProvenanceAttributes) || this;
        _this.originalGraph = originalGraph;
        _this.debugPrint = debugPrint;
        _this.reconstitutedTypes = immutable_1.Map();
        _this._lostTypeAttributes = false;
        _this._printIndent = 0;
        return _this;
    }
    BaseGraphRewriteBuilder.prototype.reconstituteType = function (t, forwardingRef) {
        return this.reconstituteTypeRef(t.typeRef, forwardingRef);
    };
    BaseGraphRewriteBuilder.prototype.reconstituteTypeRef = function (originalRef, maybeForwardingRef) {
        var maybeRef = this.lookupTypeRefs([originalRef], maybeForwardingRef);
        if (maybeRef !== undefined) {
            return maybeRef;
        }
        return this.forceReconstituteTypeRef(originalRef, maybeForwardingRef);
    };
    BaseGraphRewriteBuilder.prototype.assertTypeRefsToReconstitute = function (typeRefs, forwardingRef) {
        Support_1.assert(typeRefs.length > 0, "Must have at least one type to reconstitute");
        for (var _i = 0, typeRefs_1 = typeRefs; _i < typeRefs_1.length; _i++) {
            var originalRef = typeRefs_1[_i];
            Support_1.assert(originalRef.graph === this.originalGraph, "Trying to reconstitute a type from the wrong graph");
        }
        if (forwardingRef !== undefined) {
            Support_1.assert(forwardingRef.graph === this.typeGraph, "Trying to forward a type to the wrong graph");
        }
    };
    BaseGraphRewriteBuilder.prototype.changeDebugPrintIndent = function (delta) {
        this._printIndent += delta;
    };
    Object.defineProperty(BaseGraphRewriteBuilder.prototype, "debugPrintIndentation", {
        get: function () {
            return "  ".repeat(this._printIndent);
        },
        enumerable: true,
        configurable: true
    });
    BaseGraphRewriteBuilder.prototype.finish = function () {
        var _this = this;
        this.originalGraph.topLevels.forEach(function (t, name) {
            _this.addTopLevel(name, _this.reconstituteType(t));
        });
        return _super.prototype.finish.call(this);
    };
    BaseGraphRewriteBuilder.prototype.setLostTypeAttributes = function () {
        this._lostTypeAttributes = true;
    };
    Object.defineProperty(BaseGraphRewriteBuilder.prototype, "lostTypeAttributes", {
        get: function () {
            return this._lostTypeAttributes;
        },
        enumerable: true,
        configurable: true
    });
    return BaseGraphRewriteBuilder;
}(TypeBuilder));
exports.BaseGraphRewriteBuilder = BaseGraphRewriteBuilder;
var GraphRemapBuilder = /** @class */ (function (_super) {
    __extends(GraphRemapBuilder, _super);
    function GraphRemapBuilder(originalGraph, stringTypeMapping, alphabetizeProperties, graphHasProvenanceAttributes, _map, debugPrintRemapping) {
        var _this = _super.call(this, originalGraph, stringTypeMapping, alphabetizeProperties, graphHasProvenanceAttributes, debugPrintRemapping) || this;
        _this._map = _map;
        _this._attributeSources = immutable_1.Map();
        _map.forEach(function (target, source) {
            var maybeSources = _this._attributeSources.get(target);
            if (maybeSources === undefined) {
                maybeSources = [target];
                _this._attributeSources = _this._attributeSources.set(target, maybeSources);
            }
            maybeSources.push(source);
        });
        return _this;
    }
    GraphRemapBuilder.prototype.getMapTarget = function (tref) {
        var maybeType = this._map.get(tref.deref()[0]);
        if (maybeType === undefined)
            return tref;
        Support_1.assert(this._map.get(maybeType) === undefined, "We have a type that's remapped to a remapped type");
        return maybeType.typeRef;
    };
    GraphRemapBuilder.prototype.addForwardingIntersection = function (_forwardingRef, _tref) {
        return Support_1.panic("We can't add forwarding intersections when we're removing forwarding intersections");
    };
    GraphRemapBuilder.prototype.lookupTypeRefs = function (typeRefs, forwardingRef) {
        Support_1.assert(forwardingRef === undefined, "We can't have a forwarding ref when we remap");
        this.assertTypeRefsToReconstitute(typeRefs, forwardingRef);
        var first = this.reconstitutedTypes.get(this.getMapTarget(typeRefs[0]).getIndex());
        if (first === undefined)
            return undefined;
        for (var i = 1; i < typeRefs.length; i++) {
            var other = this.reconstitutedTypes.get(this.getMapTarget(typeRefs[i]).getIndex());
            if (first !== other)
                return undefined;
        }
        return first;
    };
    GraphRemapBuilder.prototype.forceReconstituteTypeRef = function (originalRef, maybeForwardingRef) {
        var _this = this;
        Support_1.assert(maybeForwardingRef === undefined, "We can't have a forwarding ref when we remap");
        originalRef = this.getMapTarget(originalRef);
        var _a = originalRef.deref(), originalType = _a[0], originalAttributes = _a[1];
        var attributeSources = this._attributeSources.get(originalType);
        var attributes;
        if (attributeSources === undefined) {
            attributes = originalAttributes;
        }
        else {
            attributes = Type_1.combineTypeAttributesOfTypes(attributeSources);
        }
        var index = originalRef.getIndex();
        if (this.debugPrint) {
            console.log(this.debugPrintIndentation + "reconstituting " + index);
            this.changeDebugPrintIndent(1);
        }
        var reconstituter = new TypeReconstituter(this, this.alphabetizeProperties, attributes, undefined, function (tref) {
            if (_this.debugPrint) {
                _this.changeDebugPrintIndent(-1);
                console.log(_this.debugPrintIndentation + "reconstituted " + index + " as " + tref.getIndex());
            }
            var alreadyReconstitutedType = _this.reconstitutedTypes.get(index);
            if (alreadyReconstitutedType !== undefined) {
                return Support_1.panic("We can't remap a type twice");
            }
            _this.reconstitutedTypes = _this.reconstitutedTypes.set(index, tref);
        });
        originalType.reconstitute(reconstituter);
        return reconstituter.getResult();
    };
    return GraphRemapBuilder;
}(BaseGraphRewriteBuilder));
exports.GraphRemapBuilder = GraphRemapBuilder;
var GraphRewriteBuilder = /** @class */ (function (_super) {
    __extends(GraphRewriteBuilder, _super);
    function GraphRewriteBuilder(originalGraph, stringTypeMapping, alphabetizeProperties, graphHasProvenanceAttributes, setsToReplace, debugPrintReconstitution, _replacer) {
        var _this = _super.call(this, originalGraph, stringTypeMapping, alphabetizeProperties, graphHasProvenanceAttributes, debugPrintReconstitution) || this;
        _this._replacer = _replacer;
        _this._reconstitutedUnions = immutable_1.Map();
        _this._setsToReplaceByMember = immutable_1.Map();
        var _loop_1 = function (types) {
            var set = immutable_1.Set(types);
            set.forEach(function (t) {
                var index = t.typeRef.getIndex();
                Support_1.assert(!_this._setsToReplaceByMember.has(index), "A type is member of more than one set to be replaced");
                _this._setsToReplaceByMember = _this._setsToReplaceByMember.set(index, set);
            });
        };
        for (var _i = 0, setsToReplace_1 = setsToReplace; _i < setsToReplace_1.length; _i++) {
            var types = setsToReplace_1[_i];
            _loop_1(types);
        }
        return _this;
    }
    GraphRewriteBuilder.prototype.registerUnion = function (typeRefs, reconstituted) {
        var set = immutable_1.Set(typeRefs);
        Support_1.assert(!this._reconstitutedUnions.has(set), "Cannot register reconstituted set twice");
        this._reconstitutedUnions = this._reconstitutedUnions.set(set, reconstituted);
    };
    GraphRewriteBuilder.prototype.withForwardingRef = function (maybeForwardingRef, typeCreator) {
        if (maybeForwardingRef !== undefined) {
            return typeCreator(maybeForwardingRef);
        }
        var forwardingRef = this.reserveTypeRef();
        var actualRef = typeCreator(forwardingRef);
        Support_1.assert(actualRef === forwardingRef, "Type creator didn't return its forwarding ref");
        return actualRef;
    };
    GraphRewriteBuilder.prototype.replaceSet = function (typesToReplace, maybeForwardingRef) {
        var _this = this;
        return this.withForwardingRef(maybeForwardingRef, function (forwardingRef) {
            if (_this.debugPrint) {
                console.log(_this.debugPrintIndentation + "replacing set " + typesToReplace
                    .map(function (t) { return t.typeRef.getIndex().toString(); })
                    .join(",") + " as " + forwardingRef.getIndex());
                _this.changeDebugPrintIndent(1);
            }
            typesToReplace.forEach(function (t) {
                var originalRef = t.typeRef;
                var index = originalRef.getIndex();
                _this.reconstitutedTypes = _this.reconstitutedTypes.set(index, forwardingRef);
                _this._setsToReplaceByMember = _this._setsToReplaceByMember.remove(index);
            });
            var result = _this._replacer(typesToReplace, _this, forwardingRef);
            Support_1.assert(result === forwardingRef, "The forwarding ref got lost when replacing");
            if (_this.debugPrint) {
                _this.changeDebugPrintIndent(-1);
                console.log(_this.debugPrintIndentation + "replaced set " + typesToReplace
                    .map(function (t) { return t.typeRef.getIndex().toString; })
                    .join(",") + " as " + forwardingRef.getIndex());
            }
            return result;
        });
    };
    GraphRewriteBuilder.prototype.forceReconstituteTypeRef = function (originalRef, maybeForwardingRef) {
        var _this = this;
        var _a = originalRef.deref(), originalType = _a[0], originalAttributes = _a[1];
        var index = originalRef.getIndex();
        if (this.debugPrint) {
            console.log(this.debugPrintIndentation + "reconstituting " + index);
            this.changeDebugPrintIndent(1);
        }
        var reconstituter = new TypeReconstituter(this, this.alphabetizeProperties, originalAttributes, maybeForwardingRef, function (tref) {
            if (_this.debugPrint) {
                _this.changeDebugPrintIndent(-1);
                console.log(_this.debugPrintIndentation + "reconstituted " + index + " as " + tref.getIndex());
            }
            if (maybeForwardingRef !== undefined) {
                Support_1.assert(tref === maybeForwardingRef, "We didn't pass the forwarding ref");
            }
            var alreadyReconstitutedType = _this.reconstitutedTypes.get(index);
            if (alreadyReconstitutedType === undefined) {
                _this.reconstitutedTypes = _this.reconstitutedTypes.set(index, tref);
            }
            else {
                Support_1.assert(tref.equals(alreadyReconstitutedType), "We reconstituted a type twice differently");
                if (maybeForwardingRef === undefined) {
                    return Support_1.panic("Why do we reconstitute a type twice???");
                }
            }
        });
        originalType.reconstitute(reconstituter);
        return reconstituter.getResult();
    };
    // If the union of these type refs have been, or are supposed to be, reconstituted to
    // one target type, return it.  Otherwise return undefined.
    GraphRewriteBuilder.prototype.lookupTypeRefs = function (typeRefs, forwardingRef, replaceSet) {
        if (replaceSet === void 0) { replaceSet = true; }
        this.assertTypeRefsToReconstitute(typeRefs, forwardingRef);
        // Check whether we have already reconstituted them.  That means ensuring
        // that they all have the same target type.
        var maybeRef = this.reconstitutedTypes.get(typeRefs[0].getIndex());
        if (maybeRef !== undefined && maybeRef !== forwardingRef) {
            var allEqual = true;
            for (var i = 1; i < typeRefs.length; i++) {
                if (this.reconstitutedTypes.get(typeRefs[i].getIndex()) !== maybeRef) {
                    allEqual = false;
                    break;
                }
            }
            if (allEqual) {
                return this.forwardIfNecessary(forwardingRef, maybeRef);
            }
        }
        // Has this been reconstituted as a set?
        maybeRef = this._reconstitutedUnions.get(immutable_1.Set(typeRefs));
        if (maybeRef !== undefined && maybeRef !== forwardingRef) {
            return this.forwardIfNecessary(forwardingRef, maybeRef);
        }
        // Is this set requested to be replaced?  If not, we're out of options.
        var maybeSet = this._setsToReplaceByMember.get(typeRefs[0].getIndex());
        if (maybeSet === undefined) {
            return undefined;
        }
        for (var i = 1; i < typeRefs.length; i++) {
            if (this._setsToReplaceByMember.get(typeRefs[i].getIndex()) !== maybeSet) {
                return undefined;
            }
        }
        // Yes, this set is requested to be replaced, so do it.
        if (!replaceSet)
            return undefined;
        return this.replaceSet(maybeSet, forwardingRef);
    };
    return GraphRewriteBuilder;
}(BaseGraphRewriteBuilder));
exports.GraphRewriteBuilder = GraphRewriteBuilder;
function addAttributes(accumulatorAttributes, newAttributes) {
    if (accumulatorAttributes === undefined)
        return newAttributes;
    return TypeAttributes_1.combineTypeAttributes(accumulatorAttributes, newAttributes);
}
function setAttributes(attributeMap, kind, newAttributes) {
    return attributeMap.set(kind, addAttributes(attributeMap.get(kind), newAttributes));
}
function moveAttributes(map, fromKind, toKind) {
    var fromAttributes = Support_1.defined(map.get(fromKind));
    map = map.remove(fromKind);
    return setAttributes(map, toKind, fromAttributes);
}
var UnionAccumulator = /** @class */ (function () {
    function UnionAccumulator(_conflateNumbers) {
        this._conflateNumbers = _conflateNumbers;
        this._nonStringTypeAttributes = immutable_1.OrderedMap();
        this._stringTypeAttributes = immutable_1.OrderedMap();
        this.arrayData = [];
        this.objectData = [];
        this._lostTypeAttributes = false;
        // FIXME: we're losing order here
        this.enumCaseMap = {};
        this.enumCases = [];
    }
    UnionAccumulator.prototype.have = function (kind) {
        return (this._nonStringTypeAttributes.has(kind) || this._stringTypeAttributes.has(kind));
    };
    Object.defineProperty(UnionAccumulator.prototype, "haveString", {
        get: function () {
            return this.have("string");
        },
        enumerable: true,
        configurable: true
    });
    UnionAccumulator.prototype.addAny = function (attributes) {
        this._nonStringTypeAttributes = setAttributes(this._nonStringTypeAttributes, "any", attributes);
        this._lostTypeAttributes = true;
    };
    UnionAccumulator.prototype.addNull = function (attributes) {
        this._nonStringTypeAttributes = setAttributes(this._nonStringTypeAttributes, "null", attributes);
    };
    UnionAccumulator.prototype.addBool = function (attributes) {
        this._nonStringTypeAttributes = setAttributes(this._nonStringTypeAttributes, "bool", attributes);
    };
    UnionAccumulator.prototype.addInteger = function (attributes) {
        this._nonStringTypeAttributes = setAttributes(this._nonStringTypeAttributes, "integer", attributes);
    };
    UnionAccumulator.prototype.addDouble = function (attributes) {
        this._nonStringTypeAttributes = setAttributes(this._nonStringTypeAttributes, "double", attributes);
    };
    UnionAccumulator.prototype.addStringType = function (kind, attributes) {
        if (this.have(kind)) {
            this._stringTypeAttributes = setAttributes(this._stringTypeAttributes, kind, attributes);
            return;
        }
        // string overrides all other string types, as well as enum
        if (kind === "string") {
            var oldAttributes = TypeAttributes_1.combineTypeAttributes(this._stringTypeAttributes.valueSeq().toArray());
            var newAttributes = addAttributes(oldAttributes, attributes);
            this._stringTypeAttributes = this._stringTypeAttributes.clear().set(kind, newAttributes);
            this.enumCaseMap = {};
            this.enumCases = [];
        }
        else {
            this._stringTypeAttributes = setAttributes(this._stringTypeAttributes, kind, attributes);
        }
    };
    UnionAccumulator.prototype.addArray = function (t, attributes) {
        this.arrayData.push(t);
        this._nonStringTypeAttributes = setAttributes(this._nonStringTypeAttributes, "array", attributes);
    };
    UnionAccumulator.prototype.addObject = function (t, attributes) {
        this.objectData.push(t);
        this._nonStringTypeAttributes = setAttributes(this._nonStringTypeAttributes, "object", attributes);
    };
    UnionAccumulator.prototype.addEnumCases = function (cases, attributes) {
        var _this = this;
        if (this.have("string")) {
            this.addStringType("string", attributes);
            return;
        }
        cases.forEach(function (count, s) {
            if (!Object.prototype.hasOwnProperty.call(_this.enumCaseMap, s)) {
                _this.enumCaseMap[s] = 0;
                _this.enumCases.push(s);
            }
            _this.enumCaseMap[s] += count;
        });
        this._stringTypeAttributes = setAttributes(this._stringTypeAttributes, "enum", attributes);
    };
    UnionAccumulator.prototype.addEnumCase = function (s, count, attributes) {
        this.addEnumCases(immutable_1.OrderedMap([[s, count]]), attributes);
    };
    UnionAccumulator.prototype.getMemberKinds = function () {
        var merged = this._nonStringTypeAttributes.merge(this._stringTypeAttributes);
        if (merged.isEmpty()) {
            return immutable_1.OrderedMap([["none", immutable_1.Map()]]);
        }
        if (this._nonStringTypeAttributes.has("any")) {
            Support_1.assert(this._lostTypeAttributes, "This had to be set when we added 'any'");
            var allAttributes = TypeAttributes_1.combineTypeAttributes(merged.valueSeq().toArray());
            return immutable_1.OrderedMap([["any", allAttributes]]);
        }
        if (this._conflateNumbers && this.have("integer") && this.have("double")) {
            merged = moveAttributes(merged, "integer", "double");
        }
        if (this.have("map")) {
            merged = moveAttributes(merged, "map", "class");
        }
        return merged;
    };
    Object.defineProperty(UnionAccumulator.prototype, "lostTypeAttributes", {
        get: function () {
            return this._lostTypeAttributes;
        },
        enumerable: true,
        configurable: true
    });
    return UnionAccumulator;
}());
exports.UnionAccumulator = UnionAccumulator;
var FauxUnion = /** @class */ (function () {
    function FauxUnion() {
    }
    FauxUnion.prototype.getAttributes = function () {
        return TypeAttributes_1.emptyTypeAttributes;
    };
    return FauxUnion;
}());
function attributesForTypes(types) {
    var unionsForType = immutable_1.OrderedMap();
    var typesForUnion = immutable_1.Map();
    var unions = immutable_1.OrderedSet();
    var unionsEquivalentToRoot = immutable_1.Set();
    function traverse(t, path, isEquivalentToRoot) {
        if (t instanceof Type_1.UnionType) {
            unions = unions.add(t);
            if (isEquivalentToRoot) {
                unionsEquivalentToRoot = unionsEquivalentToRoot.add(t);
            }
            path = path.add(t);
            isEquivalentToRoot = isEquivalentToRoot && t.members.size === 1;
            t.members.forEach(function (m) { return traverse(m, path, isEquivalentToRoot); });
        }
        else {
            unionsForType = unionsForType.update(t, immutable_1.Set(), function (s) { return s.union(path); });
            path.forEach(function (u) {
                typesForUnion = typesForUnion.update(u, immutable_1.Set(), function (s) { return s.add(t); });
            });
        }
    }
    var rootPath = immutable_1.Set([new FauxUnion()]);
    types.forEach(function (t) { return traverse(t, rootPath, types.size === 1); });
    var resultAttributes = unionsForType.map(function (unionForType, t) {
        var singleAncestors = unionForType.filter(function (u) { return Support_1.defined(typesForUnion.get(u)).size === 1; });
        Support_1.assert(singleAncestors.every(function (u) { return Support_1.defined(typesForUnion.get(u)).has(t); }), "We messed up bookkeeping");
        var inheritedAttributes = singleAncestors.toArray().map(function (u) { return u.getAttributes(); });
        return TypeAttributes_1.combineTypeAttributes([t.getAttributes()].concat(inheritedAttributes));
    });
    var unionAttributes = unions.toArray().map(function (u) {
        var t = typesForUnion.get(u);
        if (t !== undefined && t.size === 1) {
            return TypeAttributes_1.emptyTypeAttributes;
        }
        var attributes = u.getAttributes();
        if (unionsEquivalentToRoot.has(u)) {
            return attributes;
        }
        return TypeAttributes_1.makeTypeAttributesInferred(attributes);
    });
    return [resultAttributes, TypeAttributes_1.combineTypeAttributes(unionAttributes)];
}
// FIXME: Move this to UnifyClasses.ts?
var TypeRefUnionAccumulator = /** @class */ (function (_super) {
    __extends(TypeRefUnionAccumulator, _super);
    function TypeRefUnionAccumulator() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    // There is a method analogous to this in the IntersectionAccumulator.  It might
    // make sense to find a common interface.
    TypeRefUnionAccumulator.prototype.addType = function (t, attributes) {
        var _this = this;
        Type_1.matchTypeExhaustive(t, function (_noneType) {
            return;
        }, function (_anyType) { return _this.addAny(attributes); }, function (_nullType) { return _this.addNull(attributes); }, function (_boolType) { return _this.addBool(attributes); }, function (_integerType) { return _this.addInteger(attributes); }, function (_doubleType) { return _this.addDouble(attributes); }, function (stringType) {
            var enumCases = stringType.enumCases;
            if (enumCases === undefined) {
                _this.addStringType("string", attributes);
            }
            else {
                _this.addEnumCases(enumCases, attributes);
            }
        }, function (arrayType) { return _this.addArray(arrayType.items.typeRef, attributes); }, function (classType) { return _this.addObject(classType.typeRef, attributes); }, function (mapType) { return _this.addObject(mapType.typeRef, attributes); }, function (objectType) { return _this.addObject(objectType.typeRef, attributes); }, 
        // FIXME: We're not carrying counts, so this is not correct if we do enum
        // inference.  JSON Schema input uses this case, however, without enum
        // inference, which is fine, but still a bit ugly.
        function (enumType) { return _this.addEnumCases(enumType.cases.toOrderedMap().map(function (_) { return 1; }), attributes); }, function (_unionType) {
            return Support_1.panic("The unions should have been eliminated in attributesForTypesInUnion");
        }, function (_dateType) { return _this.addStringType("date", attributes); }, function (_timeType) { return _this.addStringType("time", attributes); }, function (_dateTimeType) { return _this.addStringType("date-time", attributes); });
    };
    TypeRefUnionAccumulator.prototype.addTypes = function (types) {
        var _this = this;
        var _a = attributesForTypes(types), attributesMap = _a[0], unionAttributes = _a[1];
        attributesMap.forEach(function (attributes, t) { return _this.addType(t, attributes); });
        return unionAttributes;
    };
    return TypeRefUnionAccumulator;
}(UnionAccumulator));
exports.TypeRefUnionAccumulator = TypeRefUnionAccumulator;
var UnionBuilder = /** @class */ (function () {
    function UnionBuilder(typeBuilder) {
        this.typeBuilder = typeBuilder;
    }
    UnionBuilder.prototype.makeTypeOfKind = function (typeProvider, kind, typeAttributes, forwardingRef) {
        switch (kind) {
            case "any":
            case "none":
            case "null":
            case "bool":
            case "double":
            case "integer":
            case "date":
            case "time":
            case "date-time":
                var t = this.typeBuilder.getPrimitiveType(kind, forwardingRef);
                this.typeBuilder.addAttributes(t, typeAttributes);
                return t;
            case "string":
                return this.typeBuilder.getStringType(typeAttributes, undefined, forwardingRef);
            case "enum":
                return this.makeEnum(typeProvider.enumCases, typeProvider.enumCaseMap, typeAttributes, forwardingRef);
            case "object":
                return this.makeObject(typeProvider.objectData, typeAttributes, forwardingRef);
            case "array":
                return this.makeArray(typeProvider.arrayData, typeAttributes, forwardingRef);
            default:
                if (kind === "union" || kind === "class" || kind === "map" || kind === "intersection") {
                    return Support_1.panic("getMemberKinds() shouldn't return " + kind);
                }
                return Support_1.assertNever(kind);
        }
    };
    UnionBuilder.prototype.buildUnion = function (typeProvider, unique, typeAttributes, forwardingRef) {
        var _this = this;
        var kinds = typeProvider.getMemberKinds();
        if (typeProvider.lostTypeAttributes) {
            this.typeBuilder.setLostTypeAttributes();
        }
        if (kinds.size === 1) {
            var _a = kinds.toArray()[0], kind = _a[0], memberAttributes = _a[1];
            var allAttributes = TypeAttributes_1.combineTypeAttributes(typeAttributes, memberAttributes);
            var t = this.makeTypeOfKind(typeProvider, kind, allAttributes, forwardingRef);
            return t;
        }
        var union = unique
            ? this.typeBuilder.getUniqueUnionType(typeAttributes, undefined, forwardingRef)
            : undefined;
        var types = [];
        kinds.forEach(function (memberAttributes, kind) {
            types.push(_this.makeTypeOfKind(typeProvider, kind, memberAttributes, undefined));
        });
        var typesSet = immutable_1.OrderedSet(types);
        if (union !== undefined) {
            this.typeBuilder.setSetOperationMembers(union, typesSet);
            return union;
        }
        else {
            return this.typeBuilder.getUnionType(typeAttributes, typesSet, forwardingRef);
        }
    };
    return UnionBuilder;
}());
exports.UnionBuilder = UnionBuilder;
